
/*****************************************************************************
 *                                                                           *
 * This FILE mx12p_lib.c                                                     *
 * is part of the MX12P library.  It is based on code copyright 1996 and     *
 * 1998 by Adimec Advanced Image Systems B.V.  Unix conversion copyright     *
 * 1999 and 2000 by Gregory L. Kreider.  All rights reserved.  See below     *
 * for distribution terms.                                                   *
 *                                                                           *
 * The code in this file provides a high level interface for programming     *
 * a PSU120/MX12P camera.  It follows the specification in the Adimec        *
 * help files.  The MX12P library stores camera information and the          *
 * connection state in global variables and uses the PSU120 library to       *
 * communicate with the camera.                                              *
 *                                                                           *
 * See libraries.doc for porting notes.                                      *
 *                                                                           *
 * This software package is provided 'as is' without any further warranties  *
 * express or implied beyond those provided for by local law.  Neither       *
 * the authors nor any distributors are liable for damages, special,         *
 * indirect, or consequential, incurred by the use or inability to use this  *
 * software, or for any claim by any other party.                            *
 *                                                                           *
 * No further restrictions are placed on the use or distribution of this     *
 * code.                                                                     *
 *                                                                           *
 *****************************************************************************/



#include <stdio.h>

#include "psu120_lib.h"
#include "mx12p_internal.h"
#include "mx12p_lib.h"


struct TUserContext_t {
  int hPSU;                 /* file descriptor for comm port */
  int Used;                 /* camera is being used */

  /* camera-specific data members */
  uchar io_reg_20;          /* requested value of 0x20 */
  uchar io_reg_21;          /* requested value of 0x21 */
  uchar io_reg_22;          /* requested value of 0x22 */
  uchar dac_reg_48;         /* requested value of 0x48 */

  /* current values of camera registers */
  uchar prev_io_reg_20;     /* current value of reg 0x20 */
  uchar prev_io_reg_21;     /* current value of reg 0x21 */
  uchar prev_io_reg_22;     /* current value of reg 0x22 */
  uchar prev_dac_reg_48;    /* current value of reg 0x48 */
  int SensorType;           /* camera type */
  int PixelClock;           /* pixel clock frequency */
};

typedef struct TUserContext_t TUserContext;

/* there are at most MAX_PORTS connections at any one time */
TUserContext UserContextTable[MAX_PORTS];


/* function prototypes */
void SInitContextEntry(TUserContext*);
int  SCamTransfer(TUserContext*, TCamParameters*);
int  SCamSaveAsDefault(TUserContext*, TCamParameters*);
int  SCamLoadDefault(TUserContext*, TCamParameters*);
int  SCamStartPoll(TUserContext*);
int  SCamPoll(TUserContext*);
int  ReadInformation(TUserContext*);
int  findQIntegrationTime(TCamParameters*);




/***************************************************************
 ***************************************************************
 * 
 *  Public Functions
 *
 ***************************************************************
 ***************************************************************/


/***************************************************************
 * 
 * CamOpen:
 *   Try to open a PSU120 channel using PSUOPEN.  
 * Returns:
 *   port number if successfully opened, or an error
 *
 ***************************************************************/

int CamOpen(int port)
{
  int fd;

  /* check that port # is valid */
  if ((port < 0) || (port >= MAX_PORTS))
	return(ERR_CAM_IPRT);
  /* check that port isn't in use (context table entry free) */
  if (UserContextTable[port].Used)
	return(ERR_CAM_RESV);
  /* open it */
  if ((fd = PSUOPEN(port)) < 0)
	return(fd);
  /* OK, it worked, so update table */
  UserContextTable[port].Used = 1;
  UserContextTable[port].hPSU = fd;
  /* initialize table entry */
  SInitContextEntry(&UserContextTable[port]);

  return(port);
}


/***************************************************************
 * 
 * CamClose:
 *   Close the PSU120 channel
 * Returns:
 *   port number if successfully closed, or an error
 *
 ***************************************************************/

int CamClose(int port)
{
  int err;

  if ((port < 0) || (port >= MAX_PORTS))
	return(ERR_CAM_IHND);
  if (!UserContextTable[port].Used)
	return(ERR_CAM_IHND);
	 
  /* OK, the comm port is valid and is in use */
  err = PSUCLOSE(UserContextTable[port].hPSU);
  UserContextTable[port].Used = 0;

  return(err);
}


/***************************************************************
 * 
 * CamTransfer:
 *   Send the camera data to the MX12P, depending on the values
 *   of the fields in TCamParameters.  Values are updated as
 *   needed.  Force update of all values by setting
 *   TCamParameters.Invalidate to something other than 0 (TRUE).
 *   TCP.Invalidate is always set 0 (FALSE) after this function.
 * Returns:
 *   ERR_CAM_IHND if invalid camera (or not in use), or the 
 *   return value of SCamTransfer
 *
 ***************************************************************/

int CamTransfer(int hCAM, TCamParameters* par)
{
  if ((hCAM < 0) || (hCAM >= MAX_PORTS))
	return(ERR_CAM_IHND);
  if (!UserContextTable[hCAM].Used)
	return(ERR_CAM_IHND);
  return(SCamTransfer(&UserContextTable[hCAM],par));
}


/***************************************************************
 * 
 * CamSaveAsDefault
 *   Send the camera data to the MX12P and save register values
 *   in NVRAM.
 * Returns:
 *   ERR_CAM_IHND if invalid camera (or not in use), or the 
 *   return value of SCamSaveAsDefault
 *
 ***************************************************************/

int CamSaveAsDefault(int hCAM, TCamParameters* par)
{
  if ((hCAM < 0) || (hCAM >= MAX_PORTS))
	return(ERR_CAM_IHND);
  if (!UserContextTable[hCAM].Used)
	return(ERR_CAM_IHND);
  return(SCamSaveAsDefault(&UserContextTable[hCAM],par));
}


/***************************************************************
 * 
 * CamLoadDefault
 *   Get the register settings from NVRAM and fills in the
 *   TCamBuf entry.
 * Returns:
 *   ERR_CAM_IHND if invalid camera (or not in use), or the 
 *   return value of SCamLoadDefault
 *
 ***************************************************************/

int CamLoadDefault(int hCAM, TCamParameters* par)
{
  if ((hCAM < 0) || (hCAM >= MAX_PORTS))
	return(ERR_CAM_IHND);
  if (!UserContextTable[hCAM].Used)
	return(ERR_CAM_IHND);
  return(SCamLoadDefault(&UserContextTable[hCAM],par));
}


/***************************************************************
 * 
 * CamStartPoll
 *   Send an ATTENTION message to the PSU120 without waiting for
 *   a reply.
 *   
 * Returns:
 *   ERR_CAM_IHND if invalid camera (or not in use), or the 
 *   return value of SCamStartPoll
 *
 ***************************************************************/

int CamStartPoll(int hCAM)
{
  if ((hCAM < 0) || (hCAM >= MAX_PORTS))
	return(ERR_CAM_IHND);
  if (!UserContextTable[hCAM].Used)
	return(ERR_CAM_IHND);
  return(SCamStartPoll(&UserContextTable[hCAM]));
}


/***************************************************************
 * 
 * CamPoll
 *   Check if an ACK byte has come back from the PSU120 (returns
 *   1 if so).  Use this function for timed polling.
 *   
 * Returns:
 *   ERR_CAM_IHND if invalid camera (or not in use), or the 
 *   return value of SCamPoll
 *
 ***************************************************************/

int CamPoll(int hCAM)
{
  if ((hCAM < 0) || (hCAM >= MAX_PORTS))
	return(ERR_CAM_IHND);
  if (!UserContextTable[hCAM].Used)
	return(ERR_CAM_IHND);
  return(SCamPoll(&UserContextTable[hCAM]));
}


/***************************************************************
 * 
 * CamGetDeviceHandle
 *   Return the file descriptor for the PSU120.
 *   
 * Returns:
 *   ERR_CAM_IHND if invalid camera (or not in use), or the 
 *   descriptor stored in the user table.
 *
 ***************************************************************/

int CamGetDeviceHandle(int hCAM)
{
  if ((hCAM < 0) || (hCAM >= MAX_PORTS))
	return(ERR_CAM_IHND);
  if (!UserContextTable[hCAM].Used)
	return(ERR_CAM_IHND);
  return(UserContextTable[hCAM].hPSU);
}


/***************************************************************
 * 
 * PrepContextTable
 *   This replaces the DLL entry function, which marks all
 *   comm ports unused.  You must call this separately from
 *   your code.
 *
 * Returns:
 *   nvt
 *   
 ***************************************************************/

void PrepContextTable(void)
{
  int i;
  for (i=0; i<MAX_PORTS; i++) {
	UserContextTable[i].Used = 0;
	SInitContextEntry(&UserContextTable[i]);
  }
}


/***************************************************************
 * 
 * CamAttention
 *   Wrapper for the PSUATTENTION command, which checks if the
 *   PSU120 is acknowledging commands.
 *
 * Returns:
 *   ERR_CAM_IHND if invalid camera (or not in use), or the 
 *   value returned from PSUATTENTION (negative if no ACK).
 *   
 ***************************************************************/

int CamAttention(int port)
{
  if ((port < 0) || (port >= MAX_PORTS))
	return(ERR_CAM_IHND);
  if (!UserContextTable[port].Used)
	return(ERR_CAM_IHND);

  return(PSUATTENTION(UserContextTable[port].hPSU));
}


/***************************************************************
 * 
 * CamVersion
 *   Writes the version string from the PSU120 into the
 *   argument.
 *
 * Returns:
 *   ERR_CAM_IHND if invalid camera (or not in use), or the 
 *   value returned from PSUVERSION.
 *   
 ***************************************************************/

int CamVersion(int port, char* verstr) 
{
  if ((port < 0) || (port >= MAX_PORTS))
	return(ERR_CAM_IHND);
  if (!UserContextTable[port].Used)
	return(ERR_CAM_IHND);

  return(PSUVERSION(UserContextTable[port].hPSU,verstr));
}
 



/***************************************************************
 ***************************************************************
 * 
 *  Internal Functions
 *
 ***************************************************************
 ***************************************************************/


/***************************************************************
 * 
 * SInitContextEntry
 *   Setup null values for all entries in the User Context
 *   
 * Returns:
 *   nvt
 *
 ***************************************************************/

void SInitContextEntry(TUserContext* Context)
{
  /* register initial values */
  Context->io_reg_20 =  0xff;
  Context->io_reg_21 =  0xff;
  Context->io_reg_22 =  0xff;
  Context->dac_reg_48 = 0;
  /* invalid values for sensor type and pixel clock */
  Context->SensorType = -1;
  Context->PixelClock = -1;
  /* previous register values (differ from initial to force update) */
  Context->prev_io_reg_20 =  0;
  Context->prev_io_reg_21 =  0;
  Context->prev_io_reg_22 =  0;
  Context->prev_dac_reg_48 = 0xff;
}


/***************************************************************
 * 
 * SCamTransfer
 *   Write the stuff out to the bus.
 *   
 * Returns:
 *   0 if successful, or error value from ReadInformation or
 *   PSUWRITE
 *
 ***************************************************************/

int SCamTransfer(TUserContext* Context, TCamParameters *par)
{
  int err;
  uchar buf[4];
  long val;

  /* Check first that the camera configuration has been read before
	 setting values that depend on b.v. sensor type */
  if ((Context->SensorType == -1) || (Context->PixelClock == -1) ||
	  (par->Invalidate == 1)) {
	if ((err = ReadInformation(Context)) < 0)
	  return(err);
  }
  par->QSensorType = Context->SensorType;
  par->QPixelClock = Context->PixelClock;

  /* Calculate integration time in [100 us] and setup configuration */
  if (par->IntegrationTime > 31) par->IntegrationTime = 31;
  if (par->IntegrationTime < 0) par->IntegrationTime = 0;
  if ((par->Mode == MODE_CONTINUOUS) && (par->IntegrationTime > 19))
	par->IntegrationTime = 19;
  par->QIntegrationTime = findQIntegrationTime(par);
  /* and also set interlacing */
  if (par->Mode == MODE_CONTROL) {
	if (par->QSensorType == SENSOR_FT18)
	  par->Interlace = 0;
  }
  else if (par->Mode != MODE_CONTINUOUS) {
	if (par->QSensorType == SENSOR_FT18)  
	  par->Interlace = 0;
  }

  /* Determine gain, in units */
  if (par->Gain > GAIN_32X)
	par->Gain = GAIN_32X;
  if (par->Gain < GAIN_1X)
	par->Gain = GAIN_1X;
  par->QGain = GainLookup[par->Gain].gain;

  /* Determine black level shift, in units */
  val = ((long) par->BlackValue * ((par->BlackFine != 0) ? 50L : 1000L))/255L;
  par->QBlackValue = (int) val;

  /* Set registers, using the macros */
  REG_OP1_SET(Context->SHD_OP1_SET, par->Option_1);
  REG_OP2_SET(Context->SHD_OP2_SET, par->Option_2);
  REG_STD_SET(Context->SHD_STD_SET, par->Standard);
  REG_TTL_SET(Context->SHD_TTL_SET, par->Interface);
  REG_SEQ_SET(Context->SHD_SEQ_SET, par->Interlace);
  REG_GAM_SET(Context->SHD_GAM_SET, par->Gamma);
  REG_MIR_SET(Context->SHD_MIR_SET, par->Mirror);
  REG_SYN_SET(Context->SHD_SYN_SET, par->Sync);
  REG_BLK_SET(Context->SHD_BLK_SET, par->Black);
  REG_FIN_SET(Context->SHD_FIN_SET, par->BlackFine);
  
  if (par->Mode > 3) par->Mode = MODE_CONTINUOUS;
  Context->SHD_MOD_VAL &= ~BIT_MOD_VAL;
  Context->SHD_MOD_VAL |= (uchar) (par->Mode << 2);

  Context->SHD_BLK_VAL = (uchar) (par->BlackValue);

  Context->SHD_GAI_VAL &= ~BIT_GAI_VAL;
  Context->SHD_GAI_VAL |= GainLookup[par->Gain].code;

  Context->SHD_INT_VAL &= ~BIT_INT_VAL;
  Context->SHD_INT_VAL |= par->IntegrationTime;

  if (par->QSensorType == SENSOR_FT12)
	REG_SES_SET( Context->SHD_SES_SET,0);
  else
	REG_SES_SET( Context->SHD_SES_SET,1);

  if (par->QPixelClock == 20)
	REG_PCSN_SET( Context->SHD_PCSN_SET,0);
  else
	REG_PCSN_SET( Context->SHD_PCSN_SET,1);

  /* Data structures set up.  Write now to PSU120 if differ from old values. */
  /* PCF8574 io-expander at I2C address 0x20 */
  if ((Context->prev_io_reg_20 != Context->io_reg_20) || par->Invalidate) {
	if ((err = PSUWRITE(Context->hPSU,0x20,1,&(Context->io_reg_20))) < 0)
	  return(err);
	Context->prev_io_reg_20 = Context->io_reg_20;
  }
  /* PCF8574 io-expander at I2C address 0x21 */
  if ((Context->prev_io_reg_21 != Context->io_reg_21) || par->Invalidate) {
	if ((err = PSUWRITE(Context->hPSU,0x21,1,&(Context->io_reg_21))) < 0)
	  return(err);
	Context->prev_io_reg_21 = Context->io_reg_21;
  }
  /* PCF8574 io-expander at I2C address 0x22 */
  if ((Context->prev_io_reg_22 != Context->io_reg_22) || par->Invalidate) {
	if ((err = PSUWRITE(Context->hPSU,0x22,1,&(Context->io_reg_22))) < 0)
	  return(err);
	Context->prev_io_reg_22 = Context->io_reg_22;
  }
  /* PCF8591 8-bit DAC at I2C address 0x48 */
  if ((Context->prev_dac_reg_48 != Context->dac_reg_48) || par->Invalidate) {
	buf[0] = 0x40;
	buf[1] = Context->dac_reg_48;
	if ((err = PSUWRITE(Context->hPSU,0x48,2,buf)) < 0)
	  return(err);
	Context->prev_dac_reg_48 = Context->dac_reg_48;
  }

  /* Always clear invalidate flag */
  par->Invalidate = 0;

  /* We've gotten here without any errors.  Congratulations! */
  return(0);
}


/***************************************************************
 * 
 * SCamPoll
 *   Check for an ACK.
 *   
 * Returns:
 *   Returns the value of the library function PSUPOLL.
 *
 ***************************************************************/

int SCamPoll(TUserContext* Context)
{
  return(PSUPOLL(Context->hPSU));
}


/***************************************************************
 * 
 * SCamStartPoll
 *   Set-up camera polling.
 *   
 * Returns:
 *   Returns the value of the library function PSUSTARTPOLL.
 *
 ***************************************************************/

int SCamStartPoll(TUserContext* Context)
{
  return(PSUSTARTPOLL(Context->hPSU));
}


/***************************************************************
 * 
 * SCamSaveAsDefault
 *   Write the register state to NVRAM after first saving the
 *   state to the PSU (also prepares the data correctly).
 *   
 * Returns:
 *   Returns the error code if the write to the PSU fails.
 *
 ***************************************************************/

int SCamSaveAsDefault(TUserContext* Context, TCamParameters* par)
{
  int err;
  uchar NVRecord[17+1];

  if ((err = SCamTransfer(Context,par)) < 0)
	return(err);

  NVRecord[ 0] = 'P';     /* identification mark */
  NVRecord[ 1] = 's';     /* 'Psu' */
  NVRecord[ 2] = 'u';

  NVRecord[ 3] = 0x20;    /* first write 0x20 */
  NVRecord[ 4] = 1;       /* message size */
  NVRecord[ 5] = Context->io_reg_20;

  NVRecord[ 6] = 0x21;    /* next write 0x21 */
  NVRecord[ 7] = 1;       /* message size */
  NVRecord[ 8] = Context->io_reg_21;

  NVRecord[ 9] = 0x22;    /* then write 0x22 */
  NVRecord[10] = 1;       /* message size */
  NVRecord[11] = Context->io_reg_22;

  NVRecord[12] = 0x48;    /* finally write 0x48 */
  NVRecord[13] = 2;       /* message size */
  NVRecord[14] = 0x40;    /* dac control code */
  NVRecord[15] = Context->dac_reg_48;

  NVRecord[16] = 0;      /* no further messages */

  return(PSUWRITENV(Context->hPSU,(uchar) 0, (uchar) 17, NVRecord));
}


/***************************************************************
 * 
 * SCamLoadDefault
 *   Actually get the register information from NVRAM and store
 *   it in the user record.
 *   
 * Returns:
 *   Returns the error code if the read from the PSU fails, or
 *   ERR_CAM_INVR if the data is inconsistant.
 *
 ***************************************************************/

int SCamLoadDefault(TUserContext* Context, TCamParameters* par)
{
  int err,i,val;
  uchar NVRecord[17+1];

  /* get the information */
  if ((err = PSUREADENV(Context->hPSU,0,17,NVRecord)) < 0) {
	return(err);
  }

  /* now check that the data is valid --- see constant data above */
  if ((NVRecord[0] != 'P') || 
	  (NVRecord[1] != 's') || 
	  (NVRecord[2] != 'u') || 
	  (NVRecord[3] != 0x20) || 
	  (NVRecord[4] != 1) ||
	  (NVRecord[6] != 0x21) ||
	  (NVRecord[7] != 1) ||
	  (NVRecord[9] != 0x22) ||
	  (NVRecord[10] != 1) ||
	  (NVRecord[12] != 0x48) ||
	  (NVRecord[13] != 2) ||
	  (NVRecord[14] != 0x40) ||
	  (NVRecord[16] != 0))
	return(ERR_CAM_INVR);

  Context->io_reg_20 = NVRecord[5];
  Context->io_reg_21 = NVRecord[8];
  Context->io_reg_22 = NVRecord[11];
  Context->dac_reg_48 = NVRecord[15];

  /* now reconstruct the camera paramters */
  par->Option_1 = VAL_OP1_SET(Context->SHD_OP1_SET);
  par->Option_2 = VAL_OP2_SET(Context->SHD_OP2_SET);
  par->Gamma = VAL_GAM_SET(Context->SHD_GAM_SET);
  par->Mirror = VAL_MIR_SET(Context->SHD_MIR_SET);
  par->Sync = VAL_SYN_SET(Context->SHD_SYN_SET);
  par->Interlace = VAL_SEQ_SET(Context->SHD_SEQ_SET);
  par->Black = VAL_BLK_SET(Context->SHD_BLK_SET);
  par->BlackFine = VAL_FIN_SET(Context->SHD_FIN_SET);
  par->IntegrationTime = Context->SHD_INT_VAL & BIT_INT_VAL;
  par->BlackValue = Context->SHD_BLK_VAL;
  par->Mode = (Context->SHD_MOD_VAL & BIT_MOD_VAL) >> 2;
  par->Standard = VAL_STD_SET(Context->SHD_STD_SET);
  par->Interface = VAL_TTL_SET(Context->SHD_TTL_SET);
  par->Gain = GAIN_32X;
  for (i=0; i<=GAIN_32X; i++) {
	if (GainLookup[i].code == (Context->SHD_GAI_VAL & BIT_GAI_VAL))
	  par->Gain = i;
  }
  /* force a total camera update on next transfer */
  par->Invalidate = 1;

  /* update calculated parameters */
  if ((err = ReadInformation(Context)) < 0)
	return(err);
  par->QSensorType = Context->SensorType;
  par->QPixelClock = Context->PixelClock;

  /* Calculate integration time in [100 us] and setup configuration */
  if (par->IntegrationTime > 31) par->IntegrationTime = 31;
  if (par->IntegrationTime < 0) par->IntegrationTime = 0;
  if ((par->Mode == MODE_CONTINUOUS) && (par->IntegrationTime > 19))
	par->IntegrationTime = 19;
  par->QIntegrationTime = findQIntegrationTime(par);

  /* Determine gain, in units */
  if (par->Gain > GAIN_32X)
	par->Gain = GAIN_32X;
  if (par->Gain < GAIN_1X)
	par->Gain = GAIN_1X;
  par->QGain = GainLookup[par->Gain].gain;

  /* Determine black level shift, in units */
  val = ((long) par->BlackValue * ((par->BlackFine != 0) ? 50L : 1000L))/255L;
  par->QBlackValue = (int) val;

  return(0);
}


/***************************************************************
 * 
 * ReadInformation
 *   Get the camera details from the PSU120.
 *   
 * Returns:
 *   Returns the error code if communication with the PSU fails.
 *
 ***************************************************************/

int ReadInformation(TUserContext *Context)
{
  int err;
  uchar buf[4];

  if ((err = PSUREAD(Context->hPSU,0x21,1,buf)) < 0)
	return(err);
  buf[1] = buf[0] | BIT_PCSN | BIT_SES;
  
  if ((err = PSUWRITE(Context->hPSU,0x21,1,&buf[1])) < 0)
	return(err);
  if ((err = PSUREAD(Context->hPSU,0x21,1,buf)) < 0)
	return(err);

  Context->SensorType = ((buf[0] & BIT_SES) > 0) ? SENSOR_FT18 : SENSOR_FT12;
  Context->PixelClock = ((buf[0] & BIT_PCSN) > 0) ? 40 : 20;

  return(0);
}


/***************************************************************
 * 
 * findQIntegrationTime
 *   Calculate the current integration time in 10^-4 seconds
 *   
 * Returns:
 *   Integration time (-1 if none)
 *
 ***************************************************************/

int findQIntegrationTime(TCamParameters* par)
{
  int index,qint;
  long val;

  switch (par->Mode) {
  case MODE_CONTINUOUS:
	index =  (par->QSensorType == SENSOR_FT18) ? 1 : 0;
	index += (par->QPixelClock == 20)          ? 2 : 0;
	index += (par->Standard == BASED_50HZ)     ? 4 : 0;
	index += (par->Interlace != 0)             ? 8 : 0;
	val = IntegrationLookup[index].LineTime;
	val *= (((par->IntegrationTime < 19) && 
			 !((par->QSensorType == SENSOR_FT18) && (par->Interlace != 0))) ?
			(par->IntegrationTime * IntegrationLookup[index].StepSize) :
			IntegrationLookup[index].MaxSize);
	qint = (int) (val/1000L);
	break;
  case MODE_CONTROL:
	qint = -1;
	break;
  default:
	qint = (200*(par->IntegrationTime + 1) *
		   ((par->QPixelClock == 20) ? 2 : 1) *
		   ((par->QSensorType == SENSOR_FT18) ? 2 : 1));
  }

  return(qint);
}





